@@ -14,6 +14,8 @@ use termion::raw::{IntoRawMode, RawTerminal};
 
 use libc::getpid;
 use time;
+use regex::Regex;
+use rand::{self, Rng};
 
 use {cost, NodeID, Coords, Node, random_fg_color, serialization, Dir, Pack};
 use plot::plot_sparkline;
@@ -21,18 +23,19 @@ use logging;
 
 pub struct Screen {
     pub max_id: u64,
-    pub drawing_root: NodeID,
     pub nodes: HashMap<NodeID, Node>,
     pub arrows: Vec<(NodeID, NodeID)>,
     pub work_path: Option<String>,
-    pub show_logs: bool,
-    pub show_meta: bool,
-    pub last_selected: Option<NodeID>,
-    pub drawing_arrow: Option<NodeID>,
+
+    // non-pub members are ephemeral
+    drawing_root: NodeID,
+    show_logs: bool,
+    last_selected: Option<NodeID>,
+    drawing_arrow: Option<NodeID>,
+    lookup: HashMap<Coords, NodeID>,
+    drawn_at: HashMap<NodeID, Coords>,
     dragging_from: Option<Coords>,
     dragging_to: Option<Coords>,
-    pub lookup: HashMap<Coords, NodeID>,
-    pub drawn_at: HashMap<NodeID, Coords>,
     stdout: Option<MouseTerminal<RawTerminal<Stdout>>>,
     // screen dimensions as detected during the current draw() cycle
     dims: Coords,
@@ -40,7 +43,7 @@ pub struct Screen {
     // where we start drawing from
     view_y: u16,
     // when we drill down then pop up, we should go to last focus, stored here
-    focus_stack: Vec<NodeID>,
+    focus_stack: Vec<(NodeID, NodeID, u16)>,
 }
 
 impl Default for Screen {
@@ -55,7 +58,6 @@ impl Default for Screen {
             lookup: HashMap::new(),
             drawn_at: HashMap::new(),
             show_logs: false,
-            show_meta: true,
             drawing_root: 0,
             stdout: None,
             dragging_from: None,
@@ -109,7 +111,7 @@ impl Screen {
                     Esc => return self.unselect().is_some(),
                     PageUp => self.scroll_up(),
                     PageDown => self.scroll_down(),
-                    Delete => self.delete_selected(),
+                    Delete => self.delete_selected(true),
                     Up => self.select_up(),
                     Down => self.select_down(),
                     Left => self.select_left(),
@@ -117,7 +119,8 @@ impl Screen {
                     Backspace => self.backspace(),
                     Char('\n') => self.create_sibling(),
                     Char('\t') => self.create_child(),
-                    Ctrl('v') => self.exec_selected(),
+                    Ctrl('n') => self.create_free_node(),
+                    Ctrl('k') => self.exec_selected(),
                     Ctrl('w') => self.drill_down(),
                     Ctrl('q') => self.pop_focus(),
                     Ctrl('f') => self.prefix_jump_prompt(),
@@ -125,11 +128,13 @@ impl Screen {
                     Ctrl('h') => self.toggle_hide_stricken(),
                     Ctrl('r') => self.add_or_remove_arrow(),
                     Ctrl('p') => self.auto_arrange(),
+                    Ctrl('z') => self.toggle_auto_arrange(),
                     Ctrl('t') => self.toggle_collapsed(),
                     Ctrl('c') => return false,
                     Ctrl('x') => self.save(),
                     Ctrl('l') => self.toggle_show_logs(),
                     Ctrl('e') => self.enter_cmd(),
+                    Ctrl('v') => self.auto_task(),
                     Char(c) => {
                         if self.last_selected.is_some() {
                             self.append(c);
@@ -161,6 +166,106 @@ impl Screen {
         true
     }
 
+    fn auto_task(&mut self) {
+        // find all leaf children of incomplete tasks
+        // if a parent is complete, the children are complete
+        // if all children are complete, but the parent isn't,
+        // we need to finish the parent
+        let mut task_roots = vec![];
+        let mut to_explore = vec![self.drawing_root];
+        while let Some(node_id) = to_explore.pop() {
+            let mut node = self.with_node(node_id, |n| n.clone()).unwrap();
+            if node.stricken {
+                // pass
+            } else if node.content.contains("#task") {
+                task_roots.push(node.id);
+            } else {
+                to_explore.append(&mut node.children);
+            }
+        }
+
+        let mut leaves = vec![];
+        while let Some(root_id) = task_roots.pop() {
+            let node = self.with_node(root_id, |n| n.clone()).unwrap();
+            let mut incomplete_children: Vec<_> = node.children
+                .iter()
+                .cloned()
+                .filter(|&c| self.with_node(c, |c| !c.stricken).unwrap())
+                .collect();
+            if incomplete_children.is_empty() {
+                leaves.push(root_id);
+            } else {
+                task_roots.append(&mut incomplete_children);
+            }
+        }
+
+        if leaves.is_empty() {
+            info!("no tasks to jump to! create some first");
+            return;
+        }
+
+        // weight based on priority of most important ancestor
+
+        let mut prio_pairs = vec![];
+        let mut total_prio = 0;
+        for &leaf in &leaves {
+            let prio = self.lineage(leaf)
+                .iter()
+                .filter_map(|&p| self.node_priority(p))
+                .max()
+                .unwrap_or(1);
+            total_prio += prio;
+            prio_pairs.push((prio, leaf));
+        }
+
+        let mut idx: usize = rand::thread_rng().gen_range(0, total_prio);
+
+        let mut choice = None;
+        for &(prio, leaf) in &prio_pairs {
+            if prio > idx {
+                choice = Some(leaf);
+                break;
+            }
+            idx -= prio;
+        }
+        let choice = choice.unwrap();
+
+        // jump to highest view where node is visible
+        let mut cursor = choice;
+        loop {
+            trace!("in auto_task loop");
+            let parent = self.parent(cursor).unwrap();
+            let collapsed = self.with_node(parent, |p| p.collapsed).unwrap();
+            cursor = parent;
+            if parent == 0 || collapsed {
+                break;
+            }
+        }
+
+        // save old location and jump
+        let old_select = self.unselect().unwrap_or(0);
+        let breadcrumb = (self.drawing_root, old_select, self.view_y);
+        self.focus_stack.push(breadcrumb);
+        self.drawing_root = cursor;
+        self.select_node(choice);
+    }
+
+    fn node_priority(&self, node_id: NodeID) -> Option<usize> {
+        lazy_static! {
+            static ref RE: Regex = Regex::new(r"#prio=(\d+)").unwrap();
+        }
+        self.with_node(node_id, |n| n.content.clone())
+            .and_then(|c| {
+                if RE.is_match(&*c) {
+                    RE.captures_iter(&*c)
+                        .nth(0)
+                        .and_then(|n| n.at(1).unwrap().parse::<usize>().ok())
+                } else {
+                    None
+                }
+            })
+    }
+
     fn single_key_prompt(&mut self, prompt: &str) -> io::Result<Key> {
         trace!("prompt({})", prompt);
         let stdin: Box<Read> = Box::new(stdin());
@@ -439,7 +544,7 @@ impl Screen {
             if let Some(selected_id) = self.last_selected {
                 if let Some(is_empty) = self.with_node(selected_id, |n| n.content.is_empty()) {
                     if is_empty {
-                        self.delete_selected();
+                        self.delete_selected(false);
                         return None;
                     } else {
                         self.last_selected.take();
@@ -533,7 +638,7 @@ impl Screen {
         }
     }
 
-    fn delete_selected(&mut self) {
+    fn delete_selected(&mut self, reselect: bool) {
         trace!("delete_selected()");
         if let Some(selected_id) = self.last_selected.take() {
             let coords = self.drawn_at.remove(&selected_id);
@@ -546,27 +651,41 @@ impl Screen {
             // remove children
             self.delete_recursive(selected_id);
             if let Some(c) = coords {
-                // need to draw here or there will be nothing to click_select below
-                self.draw();
-                self.click_select(c);
+                if reselect {
+                    // need to draw here or there will be nothing to click_select below
+                    self.draw();
+                    self.click_select(c);
+                }
             }
         }
     }
 
+    fn should_auto_arrange(&self) -> bool {
+        self.with_node(self.drawing_root, |n| n.auto_arrange).unwrap()
+    }
+
+    fn toggle_auto_arrange(&mut self) {
+        let root = self.drawing_root;
+        self.with_node_mut(root, |mut n| n.auto_arrange = !n.auto_arrange)
+            .unwrap()
+    }
+
     pub fn run(&mut self) {
         self.start_raw_mode();
         self.draw();
         let stdin = stdin();
         for c in stdin.events() {
             let evt = c.unwrap();
-            self.dims = terminal_size().unwrap();
-            let should_break = !self.handle_event(evt);
-            self.draw();
 
-            // trace!("nodes:");
-            // for (id, node) in &self.nodes {
-            // trace!("{} -> {:?} -> {}", id, node.children, node.parent_id);
-            // }
+            self.dims = terminal_size().unwrap();
+
+            let should_break = !self.handle_event(evt);
+
+            if self.should_auto_arrange() {
+                self.auto_arrange();
+            }
+
+            self.draw();
 
             if should_break {
                 self.cleanup();
@@ -624,6 +743,35 @@ impl Screen {
         }
     }
 
+    fn create_free_node(&mut self) {
+        let min_width = self.dims.0 / 3;
+        let mut y_cursor = self.view_y + 2;
+        let mut from_x = None;
+        'outer: loop {
+            trace!("in create_free_node loop");
+            // go down until we find a spot
+            // that's wide enough, then create
+            // an anchor there.
+            let mut width = 0;
+            for x in 1..self.dims.0 {
+                if self.lookup((x, y_cursor)).is_none() {
+                    width += 1;
+                    if from_x.is_none() {
+                        from_x = Some(x)
+                    }
+                    if width >= min_width {
+                        break 'outer;
+                    }
+                } else {
+                    from_x = None;
+                    width = 0;
+                }
+            }
+            y_cursor += 1;
+        }
+        self.create_anchor((from_x.unwrap(), y_cursor));
+    }
+
     fn create_anchor(&mut self, coords: Coords) {
         let root = self.drawing_root;
         let node_id = self.new_node();
@@ -632,9 +780,7 @@ impl Screen {
             node.parent_id = root;
         });
         self.with_node_mut(root, |root| root.children.push(node_id));
-        // need to draw here to populate drawn_at for select to work below
-        self.draw();
-        self.try_select(coords);
+        self.select_node(node_id);
     }
 
     fn backspace(&mut self) {
@@ -667,6 +813,20 @@ impl Screen {
         self.lookup.get(&coords)
     }
 
+    fn lineage(&self, node_id: NodeID) -> Vec<NodeID> {
+        let mut lineage = vec![node_id];
+        let mut cursor = node_id;
+        while let Some(parent) = self.parent(cursor) {
+            lineage.push(parent);
+            if parent == 0 {
+                break;
+            }
+            cursor = parent;
+        }
+        lineage.reverse();
+        lineage
+    }
+
     // returns true if a is a parent of b
     fn is_parent(&self, a: NodeID, b: NodeID) -> bool {
         trace!("is_parent({}, {})", a, b);
@@ -781,15 +941,18 @@ impl Screen {
 
     fn pop_focus(&mut self) {
         self.unselect();
-        self.drawing_root = self.focus_stack.pop().unwrap_or(0);
-        self.view_y = 0;
+        let (root, selected, view_y) = self.focus_stack.pop().unwrap_or((0, 0, 0));
+        self.drawing_root = root;
+        self.view_y = view_y;
+        self.select_node(selected);
     }
 
     fn drill_down(&mut self) {
         trace!("drill_down()");
         if let Some(selected_id) = self.unselect() {
             if selected_id != self.drawing_root {
-                self.focus_stack.push(self.drawing_root);
+                let breadcrumb = (self.drawing_root, selected_id, self.view_y);
+                self.focus_stack.push(breadcrumb);
                 self.drawing_root = selected_id;
                 self.view_y = 0;
             }
@@ -817,43 +980,51 @@ impl Screen {
         }
     }
 
-    fn scroll_to_node(&mut self, node_id: NodeID) {
+    fn scroll_to_selected(&mut self) -> bool {
+        if let Some(selected_id) = self.last_selected {
+            self.scroll_to_node(selected_id)
+        } else {
+            false
+        }
+    }
+
+    fn scroll_to_node(&mut self, node_id: NodeID) -> bool {
         if let Some(&(_, y)) = self.drawn_at(node_id) {
-            self.view_y = cmp::max(y, self.dims.1 / 2) - self.dims.1 / 2;
+            if y <= self.view_y || y > (self.view_y + self.dims.1) {
+                // move only if necessary
+                self.view_y = cmp::max(y - 1, self.dims.1 / 2) - self.dims.1 / 2;
+            }
+            true
+        } else {
+            false
         }
     }
 
     fn select_up(&mut self) {
-        if let Some(node_id) = self.find_relative_node(|cur, other| cur.1 > other.1) {
-            if node_id != 0 && !self.node_is_visible(node_id).unwrap() {
-                self.scroll_to_node(node_id);
-            }
-            self.select_node(node_id);
-        }
+        self.select_relative(|cur, other| cur.1 > other.1);
     }
 
     fn select_down(&mut self) {
-        if let Some(node_id) = self.find_relative_node(|cur, other| cur.1 < other.1) {
-            if node_id != 0 && !self.node_is_visible(node_id).unwrap() {
-                self.scroll_to_node(node_id);
-            }
-            self.select_node(node_id);
-        }
+        self.select_relative(|cur, other| cur.1 < other.1);
     }
 
     fn select_left(&mut self) {
-        if let Some(node_id) = self.find_relative_node(|cur, other| cur.0 > other.0) {
-            self.select_node(node_id);
-        }
+        self.select_relative(|cur, other| cur.0 > other.0);
     }
 
     fn select_right(&mut self) {
-        if let Some(node_id) = self.find_relative_node(|cur, other| cur.0 < other.0) {
+        self.select_relative(|cur, other| cur.0 < other.0);
+    }
+
+    fn select_relative<F>(&mut self, filter_fn: F)
+        where F: Fn(Coords, Coords) -> bool
+    {
+        if let Some(node_id) = self.find_relative_node(filter_fn) {
             self.select_node(node_id);
         }
     }
 
-    fn find_relative_node<F>(&mut self, sort_fn: F) -> Option<NodeID>
+    fn find_relative_node<F>(&mut self, filter_fn: F) -> Option<NodeID>
         where F: Fn(Coords, Coords) -> bool
     {
         let selected_id = self.last_selected.unwrap_or(0);
@@ -862,17 +1033,15 @@ impl Screen {
         let cur = self.drawn_at(selected_id).unwrap_or(&rel_def_coords);
         let (id, _) = self.drawn_at
             .iter()
-            .filter_map(|(&node, &(x, y))| {
-                let visible = y > self.view_y && y < self.view_y + self.dims.1 + 2;
-                if visible && sort_fn(*cur, (x, y)) {
-                    Some((node, cost(*cur, (x, y))))
-                } else {
-                    None
-                }
-            })
-            .fold((None, 0), |(acc_id, acc_cost), (id, cost)| {
-                if acc_cost == 0 || cost < acc_cost {
-                    (Some(id), cost)
+            .fold((None, std::u16::MAX),
+                  |(acc_id, acc_cost), (&node_id, &(x, y))| {
+                if filter_fn(*cur, (x, y)) {
+                    let cost = cost(*cur, (x, y));
+                    if cost < acc_cost {
+                        (Some(node_id), cost)
+                    } else {
+                        (acc_id, acc_cost)
+                    }
                 } else {
                     (acc_id, acc_cost)
                 }
@@ -880,11 +1049,30 @@ impl Screen {
         id
     }
 
+    fn node_cost(&self, node1: NodeID, node2: NodeID) -> Option<u16> {
+        if let Some((l1, r1)) = self.bounds_for_lookup(node1) {
+            if let Some((l2, r2)) = self.bounds_for_lookup(node2) {
+                let possibilities = vec![(l1, l2), (l1, r2), (r1, l2), (r1, r2)];
+                return possibilities.into_iter().map(|(one, two)| cost(one, two)).min();
+            }
+        }
+        None
+    }
+
     fn select_node(&mut self, node_id: NodeID) {
         trace!("select_node({})", node_id);
         self.unselect();
-        self.with_node_mut(node_id, |mut node| node.selected = true);
-        self.last_selected = Some(node_id);
+        if node_id != 0 {
+            self.with_node_mut(node_id, |mut node| node.selected = true);
+            self.last_selected = Some(node_id);
+
+            // draw() needed to make visible / scroll accurate
+            self.draw();
+
+            if !self.node_is_visible(node_id).unwrap() {
+                self.scroll_to_selected();
+            }
+        }
     }
 
     fn click_screen(&mut self, coords: Coords) {
@@ -990,9 +1178,9 @@ impl Screen {
 
     // *
     // *
-    // *
-    // * Drawing Functionality
-    // *
+    // *┌──┐
+    // *│  Drawing Functionality
+    // *└──────────────>
     // *
     // *
 
@@ -1005,7 +1193,7 @@ impl Screen {
         trace!("draw()");
         print!("{}", clear::All);
 
-        self.print_header();
+        self.draw_header();
 
         // print visible nodes
         self.draw_children_of_root();
@@ -1258,11 +1446,15 @@ impl Screen {
         print!("{}", color::Fg(color::Reset));
     }
 
-    fn print_header(&self) {
-        trace!("print_header()");
+    fn draw_header(&self) {
+        trace!("draw_header()");
         let mut header_text = self.with_node(self.drawing_root, |node| node.content.clone())
             .unwrap();
 
+        if self.should_auto_arrange() {
+            header_text.push_str(" [auto-arrange] ");
+        }
+
         let now = time::get_time().sec as u64;
         let day_in_sec = 60 * 60 * 24;
         let last_week = now - (day_in_sec * 7);
